package yapl4.interfaces;

/** This interface is used by CodeGen to issue instructions for a specific
 * target architecture. It can be viewed as a generic instruction set interface
 * for load-store architectures.
 * <p>
 * Address offsets are relative to a certain base address on the heap or in the
 * current stack frame which is determined and hided by the implementation 
 * of this interface.
 * </p><p>
 * Boolean register values are to be interpreted according to 
 * {@link #boolValue(boolean)}, which hides the actual numeric value
 * within the implementation of this interface.
 * </p>
 * @author Mario Taschwer
 * @version $Id$
 */
public interface TargetArch
{
	/** Return the number of bytes allocated by an integer data type. 
	 * @return a power of 2. 
	 */
	public int sizeofInt();
	
	/** Return the number of bytes allocated by a boolean data type.
	 * @return a power of 2. 
	 */
	public int sizeofBool();
	
	/** Return the target representation of a boolean value. */
	public int boolValue(boolean value);
	
	/** Emit a one-line comment when generating assembler code. */
	public void comment(String comment);

	/** Emit a code label with comment. 
	 * @param label		the label.
	 * @param comment	the comment, may be <code>null</code>. 
	 */
	public void writeLabel(String label, String comment);
	
	/** Allocate a given number of bytes on the heap.
	 * The allocated space will be aligned such that
	 * any of the YAPL4 primitive data types is supported.
	 * 
	 * @param bytes		the number of bytes to be allocated.
	 * @param comment	an inline comment to be emitted when
	 *                  generating assembler code.
	 * @return the offset of the lowest address of the 
	 *         allocated area on the heap. 
	 */
	public int allocHeap(int bytes, String comment);
	
	/** Allocate a given number of bytes on the stack.
	 * The allocated space will be aligned such that
	 * any of the YAPL4 primitive data types is supported.
	 * 
	 * @param bytes		the number of bytes to be allocated.
	 * @param comment	an inline comment to be emitted when
	 *                  generating assembler code.
	 * @return the offset of the lowest address of the 
	 *         allocated area on the stack. 
	 */
	public int allocStack(int bytes, String comment);
	
	/** Store the string constant on the heap.
	 * @param string	the string to be stored, enclosed in double quotes.
	 * @return the address offset on the heap.
	 */
	public int allocStringConstant(String string);
	
	/** Allocate a register.
	 * @return the register number or -1 if all registers are in use.
	 */
	public byte allocReg();
	
	/** Deallocate a register.
	 * @param reg		the register number.
	 */
	public void freeReg(byte reg);
	
	/** Issue a <em>load constant</em> (aka <em>load immediate</em>)
	 * instruction.
	 * @param reg		the destination register.
	 * @param value		the constant value (word) to be loaded.
	 */
	public void loadConst(byte reg, int value);
	
	/** Issue a <em>load word</em> instruction using an offset address.
	 * @param reg		the destination register.
	 * @param addr		the source memory offset address.
	 * @param onHeap	if <code>true</code>, <code>addr</code> is
	 *                  relative to the heap; otherwise, it is relative
	 *                  to the current stack frame.
	 */
	public void loadWord(byte reg, int addr, boolean onHeap);
	
	/** Issue a <em>load word</em> instruction using indirect addressing.
	 * <code>reg</code> will be loaded from address <code>($addrReg)+offset</code>.
	 * @param reg		the destination register.
	 * @param addrReg	the register holding the source base address.
	 * @param offset	the offset added to the value of <code>addrReg</code>
	 *                  to obtain the source memory address.
	 */
	public void loadWordReg(byte reg, byte addrReg, int offset);
	
	/** Issue a <em>load address</em> instruction.
	 * @param reg		the destination register.
	 * @param addr		the offset address to load.
	 * @param onHeap	if <code>true</code>, <code>addr</code> is
	 *                  relative to the heap; otherwise, it is relative
	 *                  to the current stack frame.
	 */
	public void loadAddress(byte reg, int addr, boolean onHeap);
	
	/** Issue a <em>store word</em> instruction using an offset address.
	 * @param reg		the source register.
	 * @param addr		the destination offset address.
	 * @param onHeap	if <code>true</code>, <code>addr</code> is
	 *                  relative to the heap; otherwise, it is relative
	 *                  to the current stack frame.
	 */
	public void storeWord(byte reg, int addr, boolean onHeap);
	
	/** Issue a <em>store word</em> instruction using indirect addressing.
	 * <code>reg</code> will be stored at address <code>($addrReg)+offset</code>.
	 * @param reg		the source register.
	 * @param addrReg	the register holding the destination base address.
	 * @param offset	the offset added to the value of <code>addrReg</code>
	 *                  to obtain the destination memory address.
	 */
	public void storeWordReg(byte reg, byte addrReg, int offset);
	
	/** Generate the procedure prolog (setup the stack frame).
	 * @param label			the assembler label to use for the procedure.
	 * @param main			set to <code>true</code> if the procedure to be entered is
	 *                      the main program (program entry point).
	 */
	public void enterProc(String label, boolean main);
	
	/** Generate the procedure epilog and return to the caller.
	 * Releases the procedure's stack frame.
	 * @param label			the assembler label to use at epilog entry.
	 */
	public void exitProc(String label);
	
	/** Return from procedure by jumping to the procedure epilog.
	 * @param label		the procedure epilog label; must match the label given
	 *                  to {@link #exitProc(String)}.
	 * @param reg		the register holding the return value. Set this to -1
	 *                  if the procedure does not return a value.
	 */
	public void returnFromProc(String label, byte reg);
	
	/** Prepare stack frame for procedure call.
	 * DO NOT allocate space on the stack between calling
	 * this method and calling {@link #callProc(byte, String)}.
	 * @param numArgs		the number of procedure arguments.
	 */
	public void prepareProcCall(int numArgs);
	
	/** Pass an argument residing in a register to a procedure.
	 * Needs to be called after {@link #prepareProcCall(int)}.
	 * @param arg		the argument number, starting at 0.
	 * @param reg		the register number where the argument
	 * 					value resides.
	 */
	public void passArg(int arg, byte reg);
	
	/** Generate a procedure call.
	 * {@link #prepareProcCall(int)} MUST be called before.
	 * Procedure arguments (if any) need to be passed using
	 * {@link #passArg(int, byte)} before.
	 * @param reg			register number for storing the
	 * 						return value. Set this parameter to
	 *                      -1, if the procedure does not return
	 *                      a value.
	 * @param name			the procedure name (label); must match
	 *                      the label given to {@link #enterProc(String, boolean)}.
	 */
	public void callProc(byte reg, String name);
	
	/** Return the stack address offset for a given procedure parameter index,
	 * relative to the procedure's stack frame.
	 * @param index		the parameter index, starts at 0.
	 */
	public int paramOffset(int index);
	
	/** Return a register initialized with zero. */
	public byte zeroReg();
	
	/** Add integer registers.<br>
	 * <code>regDest = regX + regY</code><br>
	 * The destination register may be equal to one of the source registers.
	 * @param regDest		destination register number.
	 * @param regX			register number of first source operand.
	 * @param regY			register number of second source operand.
	 */
	public void add(byte regDest, byte regX, byte regY);

	/** Subtract integer registers.<br>
	 * <code>regDest = regX - regY</code><br>
	 * The destination register may be equal to one of the source registers.
	 * @param regDest		destination register number.
	 * @param regX			register number of first source operand.
	 * @param regY			register number of second source operand.
	 */
	public void sub(byte regDest, byte regX, byte regY);

	/** Multiply integer registers.<br>
	 * <code>regDest = regX * regY</code><br>
	 * The destination register may be equal to one of the source registers.
	 * @param regDest		destination register number.
	 * @param regX			register number of first source operand.
	 * @param regY			register number of second source operand.
	 */
	public void mul(byte regDest, byte regX, byte regY);

	/** Divide integer registers.<br>
	 * <code>regDest = regX / regY</code><br>
	 * The destination register may be equal to one of the source registers.
	 * @param regDest		destination register number.
	 * @param regX			register number of first source operand.
	 * @param regY			register number of second source operand.
	 */
	public void div(byte regDest, byte regX, byte regY);
	
	/** Set boolean register value by <code>less than</code> comparison.<br>
	 * <code>regDest = (regX < regY)</code><br>
	 * The destination register may be equal to one of the source registers.
	 * @param regDest		destination register number.
	 * @param regX			register number of first source operand.
	 * @param regY			register number of second source operand.
	 */
	public void isLess(byte regDest, byte regX, byte regY);
	
	/** Set boolean register value by <code>is equal</code> comparison.<br>
	 * <code>regDest = (regX == regY)</code><br>
	 * The destination register may be equal to one of the source registers.
	 * @param regDest		destination register number.
	 * @param regX			register number of first source operand.
	 * @param regY			register number of second source operand.
	 */
	public void isEqual(byte regDest, byte regX, byte regY);
	
	/** Logical NOT of a boolean register value.<br>
	 * <code>regDest = NOT regSrc</code><br>
	 * The destination register may be equal to the source register.
	 * @param regDest		destination register number.
	 * @param regSrc		source register number.
	 */
	public void not(byte regDest, byte regSrc);

	/** Logical AND of boolean register values.<br>
	 * <code>regDest = regX AND regY</code><br>
	 * The destination register may be equal to one of the source registers.
	 * @param regDest		destination register number.
	 * @param regX			register number of first source operand.
	 * @param regY			register number of second source operand.
	 */
	public void and(byte regDest, byte regX, byte regY);
	
	/** Logical OR of boolean register values.<br>
	 * <code>regDest = regX OR regY</code><br>
	 * The destination register may be equal to one of the source registers.
	 * @param regDest		destination register number.
	 * @param regX			register number of first source operand.
	 * @param regY			register number of second source operand.
	 */
	public void or(byte regDest, byte regX, byte regY);
	
	/** Emit branch instruction jumping to <code>label</code>
	 * if a register represents a given boolean value.
	 * @param reg		the register representing a boolean value.
	 * @param value		the boolean value to compare with <code>reg</code>.
	 * @param label		the label to jump to. 
	 */
	public void branchIf(byte reg, boolean value, String label);
	
	/** Emit unconditional jump to <code>label</code>. */
	public void jump(String label);
}